class EnodoError(Exception):
    def __init__(self, *args, errdata=None):
        if isinstance(errdata, dict):
            args = (errdata['error_msg'],)
            self.error_code = errdata['error_code']
        super().__init__(*args)


class EnodoConnectionError(EnodoError):
    pass



class EnodoModuleArgument(dict):

    def __init__(self, name, required, description, job_types):
        super(EnodoModuleArgument, self).__init__({
            "name": name,
            "required": required,
            "description": description,
            "job_types": job_types})

    @property
    def name(self):
        return self.get("name")

    @property
    def required(self):
        return self.get("required")

    @property
    def description(self):
        return self.get("description")

    @property
    def job_types(self):
        return self.get("job_types")


class EnodoModule(dict):
    def __init__(
            self, name, module_arguments, supported_jobs=[],
            job_load_weight={}):
        """
        :param name:
        :param module_arguments:  EnodoModuleArgument
        """
        super(EnodoModule, self).__init__({
            "name": name,
            "module_arguments": module_arguments,
            "supported_jobs": supported_jobs,
            "job_load_weight": job_load_weight
        })

    @property
    def name(self):
        return self.get("name")

    @property
    def module_arguments(self):
        return self.get("module_arguments")

    @property
    def supported_jobs(self):
        return self.get("supported_jobs")

    @property
    def job_load_weight(self):
        return self.get("job_load_weight")

    def support_job_type(self, job_type):
        return job_type in self.supported_jobs

    def get_job_load_weight(self, job_type):
        return self.job_load_weight.get(job_type)



import json
import logging
from abc import abstractmethod


class EnodoJobDataModel():

    def __init__(self, **kwargs):
        self._dict_values = kwargs
        if not self.validate():
            raise Exception("invalid data for packaga data")

    def validate(self):
        if self.required_fields is not None:
            if "errors" in self._dict_values.keys():
                return True
            for key in self.required_fields:
                if key not in self._dict_values.keys():
                    logging.error(f"Missing '{key}' in enodo "
                                  "job data model data")
                    return False
        return "model_type" in self._dict_values.keys() and \
            self.validate_data(self._dict_values)

    @property
    @abstractmethod
    def required_fields(self):
        """ return list of required fields """

    def validate_data(self, data):
        """ validate data """
        return True

    def get(self, key):
        return self._dict_values.get(key)

    def _children_to_dict(self):
        r = {}
        for key, child in self._dict_values.items():
            r[key] = child

        return r

    def serialize(self):
        return json.dumps(self._children_to_dict())

    @classmethod
    def unserialize(cls, json_data):
        data = json.loads(json_data)
        model_type = data.get("model_type")

        if model_type == "forecast_response":
            return EnodoForecastJobResponseDataModel(**data)
        elif model_type == "anomaly_response":
            return EnodoDetectAnomaliesJobResponseDataModel(**data)
        elif model_type == "base_response":
            return EnodoBaseAnalysisJobResponseDataModel(**data)
        elif model_type == "static_rules_response":
            return EnodoStaticRulesJobResponseDataModel(**data)
        elif model_type == "job_request":
            return EnodoJobRequestDataModel(**data)

        return None

    @classmethod
    def validate_by_job_type(cls, data, job_type):

        try:
            if job_type == "job_forecast":
                return EnodoForecastJobResponseDataModel(**data)
            elif job_type == "job_anomaly_detect":
                return EnodoDetectAnomaliesJobResponseDataModel(**data)
            elif job_type == "job_base_analysis":
                return EnodoBaseAnalysisJobResponseDataModel(**data)
            elif job_type == "job_static_rules":
                return EnodoStaticRulesJobResponseDataModel(**data)
            elif job_type == "job_request":
                return EnodoJobRequestDataModel(**data)
        except Exception as _:
            return False

        return True


class EnodoJobRequestDataModel(EnodoJobDataModel):

    def __init__(self, **kwargs):
        kwargs['model_type'] = "job_request"
        super().__init__(**kwargs)

    @property
    def required_fields(self):
        return [
            "job_id",
            "series_name",
            "job_config",
            "series_config",
            "series_state"
        ]

    # TODO add optional fields for explicity
    # Optional: required_job_config


class EnodoForecastJobResponseDataModel(EnodoJobDataModel):

    def __init__(self, **kwargs):
        kwargs['model_type'] = "forecast_response"
        super().__init__(**kwargs)

    @property
    def required_fields(self):
        return [
            "successful",
            "data"
        ]


class EnodoDetectAnomaliesJobResponseDataModel(EnodoJobDataModel):

    def __init__(self, **kwargs):
        kwargs['model_type'] = "anomaly_response"
        super().__init__(**kwargs)

    @property
    def required_fields(self):
        return [
            "successful",
            "data"
        ]


class EnodoBaseAnalysisJobResponseDataModel(EnodoJobDataModel):

    def __init__(self, **kwargs):
        kwargs['model_type'] = "base_response"
        super().__init__(**kwargs)

    @property
    def required_fields(self):
        return [
            "successful",
            "characteristics",
            "health"
        ]


class EnodoStaticRulesJobResponseDataModel(EnodoJobDataModel):

    def __init__(self, **kwargs):
        kwargs['model_type'] = "static_rules_response"
        super().__init__(**kwargs)

    @property
    def required_fields(self):
        return [
            "successful",
            "data"
        ]

    def validate_data(self, data):
        if not isinstance(data['data'], list):
            return False

        for failed_check in data['data']:
            if not isinstance(failed_check, list):
                return False

        return True



import asyncio
import datetime
import errno
import fcntl
import os
import uuid
import logging

import qpack

from enodo.exceptions import EnodoConnectionError
from .protocol.package import *
from .version import __version__ as VERSION


class Client:

    def __init__(self, loop, hostname, port, client_type, token,
                 identity_file_path=None, heartbeat_interval=5,
                 client_version=VERSION):
        self.loop = loop
        self._hostname = hostname
        self._port = port
        self._heartbeat_interval = heartbeat_interval
        self._client_type = client_type
        self._client_version = client_version
        self._id = uuid.uuid4().hex
        if identity_file_path is not None:
            enodo_id = self.read_enodo_id(identity_file_path)
            if enodo_id is None or enodo_id == "":
                self.write_enodo_id(identity_file_path, self._id)
            else:
                self._id = enodo_id

        self._token = token
        self._messages = {}

        self._last_heartbeat_send = datetime.datetime.now()
        self._updates_on_heartbeat = []
        self._cbs = None
        self._handshake_data_cb = None
        self._sock = None
        self._running = True
        self._connected = False

    async def setup(self, cbs=None, handshake_cb=None):
        await self._connect()

        self._cbs = cbs
        if cbs is None:
            self._cbs = {}
        if handshake_cb is not None:
            self._handshake_data_cb = handshake_cb

        await self._handshake()

    def read_enodo_id(self, path):
        enodo_id = os.getenv(f"ENODO_ID")
        if not (enodo_id is None or enodo_id == ""):
            return enodo_id
        if not os.path.exists(path):
            return None
        with open(path, 'r') as f:
            enodo_id = f.read()
        return enodo_id

    def write_enodo_id(self, path, enodo_id):
        with open(path, 'w') as f:
            f.write(enodo_id)
        return True

    async def _connect(self):
        while not self._connected and self._running:
            logging.info("Trying to connect")
            try:
                self._sock = socket.socket(
                    socket.AF_INET, socket.SOCK_STREAM)
                self._sock.connect((self._hostname, self._port))
            except Exception as e:
                logging.warning(f"Cannot connect, {str(e)}")
                logging.info("Retrying in 5")
                await asyncio.sleep(5)
            else:
                logging.info("Connected")
                self._connected = True
                fcntl.fcntl(self._sock, fcntl.F_SETFL, os.O_NONBLOCK)

    async def run(self):
        while self._running:
            diff = datetime.datetime.now() - self._last_heartbeat_send
            if diff.total_seconds() > int(
                    self._heartbeat_interval):
                await self._send_heartbeat()

            await self._read_from_socket()
            await asyncio.sleep(1)

    async def close(self):
        logging.info('Closing the socket')
        self._running = False
        self._sock.close()

    async def _read_from_socket(self):
        try:
            header = self._sock.recv(PACKET_HEADER_LEN)
        except socket.error as e:
            err = e.args[0]
            if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                pass
            else:
                raise EnodoConnectionError
        else:
            await self._read_message(header)

    async def _read_message(self, header):
        packet_type, packet_id, data = await read_packet(self._sock, header)

        if len(data):
            data = qpack.unpackb(data, decode='utf-8')

        if packet_type == 0:
            logging.warning("Connection lost, trying to reconnect")
            self._connected = False
            try:
                await self.setup(self._cbs)
            except Exception as e:
                logging.error('Error while trying to setup client')
                logging.debug(f'Correspondig error: {str(e)}')
                await asyncio.sleep(5)
        elif packet_type == HANDSHAKE_OK:
            logging.info(f'Hands shaked with hub')
        elif packet_type == HANDSHAKE_FAIL:
            logging.warning(f'Hub does not want to shake hands')
        elif packet_type == HEARTBEAT:
            logging.debug(f'Heartbeat back from hub')
        elif packet_type == RESPONSE_OK:
            logging.debug(f'Hub received update correctly')
        elif packet_type == UNKNOWN_CLIENT:
            logging.error(f'Hub does not recognize us')
            await self._handshake()
        else:
            if packet_type in self._cbs.keys():
                await self._cbs.get(packet_type)(data)
            else:
                logging.error(
                    f'Message type not implemented: {packet_type}')

    async def _send_message(self, length, message_type, data):
        header = create_header(length, message_type)

        logging.debug(f"Sending type: {message_type}")
        self._sock.send(header + data)

    async def send_message(self, body, message_type, use_qpack=True):
        if not self._connected:
            return False
        if use_qpack:
            body = qpack.packb(body)
        await self._send_message(len(body), message_type, body)

    async def _handshake(self):
        data = {
            'client_id': str(self._id),
            'client_type': self._client_type,
            'token': self._token,
            'version': self._client_version}
        if self._handshake_data_cb is not None:
            handshake_data = await self._handshake_data_cb()
            data = {**data, **handshake_data}
        data = qpack.packb(data)
        await self._send_message(len(data), HANDSHAKE, data)
        self._last_heartbeat_send = datetime.datetime.now()

    async def _send_heartbeat(self):
        logging.debug('Sending heartbeat to hub')
        id_encoded = qpack.packb(self._id)
        await self._send_message(len(id_encoded), HEARTBEAT, id_encoded)
        self._last_heartbeat_send = datetime.datetime.now()



JOB_TYPE_FORECAST_SERIES = "job_forecast"
JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES = "job_anomaly_detect"
JOB_TYPE_BASE_SERIES_ANALYSIS = "job_base_analysis"
JOB_TYPE_STATIC_RULES = "job_static_rules"

JOB_TYPES = [
    JOB_TYPE_BASE_SERIES_ANALYSIS,
    JOB_TYPE_FORECAST_SERIES,
    JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES,
    JOB_TYPE_STATIC_RULES
]

JOB_STATUS_NONE = None
JOB_STATUS_OPEN = 1
JOB_STATUS_PENDING = 2
JOB_STATUS_DONE = 3
JOB_STATUS_FAILED = 4



import struct
import qpack


class Package:

    __slots__ = ('pid', 'length', 'tipe', 'checkbit', 'data')

    struct_datapackage = struct.Struct('<IHBB')

    def __init__(self, barray):
        self.length, self.pid, self.tipe, self.checkbit = \
            self.__class__.struct_datapackage.unpack_from(barray, offset=0)
        self.length += self.__class__.struct_datapackage.size
        self.data = None

    def extract_data_from(self, barray):
        try:
            self.data = qpack.unpackb(
                barray[self.__class__.struct_datapackage.size:self.length],
                decode='utf-8')
        finally:
            del barray[:self.length]



WORKER_MODE_GLOBAL = "global"
WORKER_MODE_DEDICATED_JOB_TYPE = "dedicated_job_type"
WORKER_MODE_DEDICATED_SERIES = "dedicated_series"
WORKER_MODES = [
    WORKER_MODE_GLOBAL,
    WORKER_MODE_DEDICATED_JOB_TYPE,
    WORKER_MODE_DEDICATED_SERIES
]


class WorkerConfigModel(dict):
    def __init__(self, mode,
                 dedicated_job_type=None, dedicated_series_name=None):
        if mode not in WORKER_MODES:
            raise Exception("Invalid worker mode")

        if mode == WORKER_MODE_DEDICATED_JOB_TYPE and \
                dedicated_job_type is None:
            raise Exception("Invalid worker config")

        if mode == WORKER_MODE_DEDICATED_SERIES and \
                dedicated_series_name is None:
            raise Exception("Invalid worker config")

        super(WorkerConfigModel, self).__init__({
            "mode": mode,
            "dedicated_job_type": dedicated_job_type,
            'dedicated_series_name': dedicated_series_name
        })

    @property
    def mode(self):
        return self.get("mode")

    @mode.setter
    def mode(self, value):
        self["mode"] = value

    @property
    def dedicated_job_type(self):
        return self.get("dedicated_job_type")

    @dedicated_job_type.setter
    def dedicated_job_type(self, value):
        self["dedicated_job_type"] = value

    @property
    def dedicated_series_name(self):
        return self.get("dedicated_series_name")

    @dedicated_series_name.setter
    def dedicated_series_name(self, value):
        self["dedicated_series_name"] = value



import functools
from asyncio import events, ensure_future, futures
from asyncio.tasks import _release_waiter, _cancel_and_wait


async def _waiter_check_cancel(cb):
    while True:
        if await cb():
            return True
        import time
        # time.sleep(1)
        import asyncio
        await asyncio.sleep(1)


async def wait_for_with_cancel(fut, timeout, *, loop=None, check_cancel=None):
    """Wait for the single Future or coroutine to complete, with timeout.

    Coroutine will be wrapped in Task.

    Returns result of the Future or coroutine.  When a timeout occurs,
    it cancels the task and raises TimeoutError.  To avoid the task
    cancellation, wrap it in shield().

    If the wait is cancelled, the task is also cancelled.

    This function is a coroutine.
    """
    if loop is None:
        loop = events.get_event_loop()

    if timeout is None:
        return await fut

    if timeout <= 0:
        fut = ensure_future(fut, loop=loop)

        if fut.done():
            return fut.result()

        fut.cancel()
        raise futures.TimeoutError()

    if check_cancel is None:
        waiter = loop.create_future()
    else:
        waiter = _waiter_check_cancel(check_cancel)

    timeout_handle = loop.call_later(timeout, _release_waiter, waiter)
    cb = functools.partial(_release_waiter, waiter)

    fut = ensure_future(fut, loop=loop)
    fut.add_done_callback(cb)

    try:
        # wait until the future completes or the timeout
        try:
            await waiter
        except futures.CancelledError:
            fut.remove_done_callback(cb)
            fut.cancel()
            raise

        if fut.done():
            return fut.result()
        else:
            fut.remove_done_callback(cb)
            # We must ensure that the task is not running
            # after wait_for() returns.
            # See https://bugs.python.org/issue32751
            await _cancel_and_wait(fut, loop=loop)
            raise futures.TimeoutError()
    finally:
        timeout_handle.cancel()



import uuid

from enodo.jobs import (
    JOB_TYPES,
    JOB_STATUS_NONE,
)


class SeriesJobConfigModel(dict):

    def __init__(self, module, job_type, job_schedule_type,
                 job_schedule, module_params, max_n_points=None,
                 activated=True, config_name=None, silenced=False,
                 requires_job=None):

        if not isinstance(activated, bool):
            raise Exception(
                "Invalid series job config, activated property must be a bool")

        if not isinstance(module, str):
            raise Exception(
                "Invalid series job config, module property must be a string")

        if job_type not in JOB_TYPES:
            raise Exception(
                "Invalid series job config, unknown job_type")

        if not isinstance(job_schedule_type, str):
            raise Exception(
                "Invalid series job config, "
                "job_schedule_type property must be a string")

        if job_schedule_type not in ['N', 'TS']:
            raise Exception(
                "Invalid series job config, "
                "job_schedule_type property be one of: ['N', 'TS']")

        if not isinstance(job_schedule, int):
            raise Exception(
                "Invalid series job config, "
                "job_schedule property must be an integer")

        if not isinstance(module_params, dict):
            raise Exception(
                "Invalid series job config, "
                "module_params property must be a dict")

        if max_n_points is not None and not isinstance(max_n_points, int):
            raise Exception(
                "Invalid series job config, "
                "max_n_points property must be an integer")

        if not isinstance(silenced, bool):
            raise Exception(
                "Invalid series job config, "
                "silenced property must be a bool")

        if config_name is None:
            config_name = str(uuid.uuid4())

        super(SeriesJobConfigModel, self).__init__({
            "activated": activated,
            "module": module,
            "job_type": job_type,
            "job_schedule_type": job_schedule_type,
            "job_schedule": job_schedule,
            "max_n_points": max_n_points,
            "module_params": module_params,
            "config_name": config_name,
            "silenced": silenced,
            "requires_job": requires_job})

    @property
    def activated(self):
        return self.get("activated")

    @property
    def module(self):
        return self.get("module")

    @property
    def job_type(self):
        return self.get("job_type")

    @property
    def job_schedule_type(self):
        return self.get("job_schedule_type")

    @property
    def job_schedule(self):
        return self.get("job_schedule")

    @property
    def max_n_points(self):
        return self.get("max_n_points")

    @property
    def module_params(self):
        return self.get("module_params")

    @property
    def config_name(self):
        return self.get("config_name")

    @property
    def silenced(self):
        return self.get("silenced")

    @property
    def requires_job(self):
        return self.get("requires_job")


class SeriesConfigModel(dict):

    def __init__(
            self, job_config, min_data_points=None, realtime=False):
        """
        Create new Series Config
        :param job_config: dict of job(key) and config(value)
        :param min_data_points: int value of min points before it will be
            analysed or used in a job
        :param realtime: boolean if series should be analysed in realtime with
            datapoint updates
        :return:
        """

        if not isinstance(job_config, list):
            raise Exception(
                "Invalid series config, job_config property must be a list")

        _job_config_list = []
        for job in job_config:
            jmc = SeriesJobConfigModel(**job)
            _job_config_list.append(jmc)

        if not isinstance(min_data_points, int):
            raise Exception(
                "Invalid series config, "
                "min_data_points property must be an integer")

        if not isinstance(realtime, bool):
            raise Exception(
                "Invalid series config, realtime property must be a bool")

        super(SeriesConfigModel, self).__init__({
            "job_config": _job_config_list,
            "min_data_points": min_data_points,
            "realtime": realtime})

    @property
    def job_config(self):
        _job_config = {}
        for jmc in self['job_config']:
            _job_config[jmc.config_name] = jmc
        return _job_config

    @property
    def min_data_points(self):
        return self.get("min_data_points")

    @property
    def realtime(self):
        return self.get("realtime")

    def get_config_for_job_type(self, job_type, first_only=True):
        r = []
        for job in self['job_config']:
            if job.job_type == job_type:
                r.append(job)

        if first_only:
            return r[0] if len(r) > 0 else None
        return r

    def get_config_for_job(self, job_config_name):
        return self.job_config.get(job_config_name)


class JobSchedule(dict):

    def __init__(self, schedule=None):

        if schedule is None:
            schedule = {}
        else:
            if not isinstance(schedule, dict):
                raise Exception("Invalid series job schedule")
            for job_config_name, schedule_item in schedule.items():
                if "value" not in schedule_item or "type" not in schedule_item:
                    raise Exception("Invalid series job schedule")
        super(JobSchedule, self).__init__(schedule)

    def get_job_schedule(self, job_config_name):
        return self.get(job_config_name, None)

    def set_job_schedule(self, job_config_name, value):
        if job_config_name not in self:
            self[job_config_name] = {}

        self[job_config_name] = value


class JobStatuses(dict):

    def __init__(self, statuses=None):

        if statuses is None:
            statuses = {}
        else:
            if not isinstance(statuses, dict):
                raise Exception("Invalid series job statuses")

        super(JobStatuses, self).__init__(statuses)

    def get_job_status(self, job_config_name):
        return self.get(job_config_name, JOB_STATUS_NONE)

    def set_job_status(self, job_config_name, value):
        self[job_config_name] = value


class JobCheckStatuses(dict):

    def __init__(self, statuses=None):

        if statuses is None:
            statuses = {}
        else:
            if not isinstance(statuses, dict):
                raise Exception("Invalid series job check statuses")

        super(JobCheckStatuses, self).__init__(statuses)

    def get_job_check_status(self, job_config_name):
        return self.get(job_config_name, "")

    def set_job_check_status(self, job_config_name, status_mesage):
        self[job_config_name] = status_mesage


class SeriesState(dict):

    def __init__(
            self,
            datapoint_count=None,
            health=None,
            interval=None,
            job_schedule=None,
            job_check_statuses=None,
            job_statuses=None):

        job_schedule = JobSchedule(job_schedule)
        job_statuses = JobStatuses(job_statuses)
        job_check_statuses = JobCheckStatuses(job_check_statuses)

        super(SeriesState, self).__init__({
            "datapoint_count": datapoint_count,
            "health": health,
            "interval": interval,
            "job_schedule": job_schedule,
            "job_statuses": job_statuses,
            "job_check_statuses": job_check_statuses
        })

    @property
    def datapoint_count(self):
        return self.get("datapoint_count")

    @datapoint_count.setter
    def datapoint_count(self, value):
        self["datapoint_count"] = value

    @property
    def health(self):
        return self.get("health")

    @health.setter
    def health(self, value):
        self["health"] = value

    @property
    def interval(self):
        return self.get("interval")

    @interval.setter
    def interval(self, value):
        self["interval"] = value

    @property
    def job_schedule(self):
        return self.get("job_schedule")

    @job_schedule.setter
    def job_schedule(self, value):
        self["job_schedule"] = value

    @property
    def job_statuses(self):
        return self.get("job_statuses")

    @job_statuses.setter
    def job_statuses(self, value):
        self["job_statuses"] = value

    @property
    def job_check_statuses(self):
        return self.get("job_check_statuses")

    @job_check_statuses.setter
    def job_check_statuses(self, value):
        self["job_check_statuses"] = value

    def get_job_status(self, job_config_name):
        return self['job_statuses'].get_job_status(job_config_name)

    def set_job_status(self, job_config_name, value):
        self['job_statuses'].set_job_status(job_config_name, value)

    def get_all_job_schedules(self):
        return self['job_schedule']

    def get_job_schedule(self, job_config_name):
        return self['job_schedule'].get_job_schedule(job_config_name)

    def set_job_schedule(self, job_config_name, value):
        self['job_schedule'].set_job_schedule(job_config_name, value)

    def get_job_check_status(self, job_config_name):
        return self['job_check_statuses'].get_job_check_status(
            job_config_name)

    def set_job_check_status(self, job_config_name, status_message):
        self['job_check_statuses'].set_job_check_status(
            job_config_name,
            status_message)



import asyncio
import logging
import traceback

from enodo.jobs import JOB_TYPE_FORECAST_SERIES, \
    JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES, \
    JOB_TYPE_BASE_SERIES_ANALYSIS, JOB_TYPE_STATIC_RULES
from enodo.model.config.series import SeriesJobConfigModel
from enodo.protocol.packagedata import EnodoJobDataModel

from .lib.siridb.siridb import SiriDB, config_equals


class Analyser:
    _analyser_queue = None
    _busy = None
    _siridb_data_client = None
    _siridb_output_client = None
    _shutdown = None
    _current_future = None

    def __init__(
            self, queue, siridb_data, siridb_output, modules):
        self._siridb_data_client = self._siridb_output_client = SiriDB(
            siridb_data['user'],
            siridb_data['password'],
            siridb_data['database'],
            siridb_data['host'],
            siridb_data['port'])
        if not config_equals(siridb_data, siridb_output):
            self._siridb_output_client = SiriDB(
                siridb_output['user'],
                siridb_output['password'],
                siridb_output['database'],
                siridb_output['host'],
                siridb_output['port'])
        self._analyser_queue = queue
        self._modules = modules

    async def query_siridb(self, query, output=False):
        if output:
            return await self._siridb_output_client.run_query(query)
        return await self._siridb_data_client.run_query(query)

    async def execute_job(self, job_data):
        series_name = job_data.get("series_name")
        series_state = job_data.get("series_state")
        job_config = SeriesJobConfigModel(**job_data.get('job_config'))
        max_n_points = job_config.get('max_n_points')
        time_unit = job_data.get("time_unit")
        job_type = job_config.job_type
        timeval = None

        time_unit = (await self._siridb_data_client.run_query(
            "show time_precision"))["data"][0]["value"]

        series_data = await self._siridb_data_client.query_series_data(
            series_name, since=timeval)

        # if max_n_points is not None and \
        #         series_state.get("interval") is not None and \
        #         time_unit is not None:
        #     timeval = max_n_points * int(series_state.get("interval"))
        #     if time_unit == "ms":
        #         timeval = int(timeval / 1000)
        #     timeval = int(timeval / 60)  # In minutes

        if series_name not in series_data:
            return self._analyser_queue.put(
                {'name': series_name,
                 'error': 'Cannot find series data'})
        # TODO: use tail function
        dataset = series_data[series_name][-1000:]
        parameters = job_config.module_params

        module_class = self._modules.get(job_config.module)

        if module_class is not None:
            module = module_class(dataset, parameters,
                                  series_name, self.query_siridb)

            if job_type == JOB_TYPE_BASE_SERIES_ANALYSIS:
                await self._analyse_series(series_name, module)
            elif job_type == JOB_TYPE_STATIC_RULES:
                await self._check_static_rules(series_name, module)
            elif job_type == JOB_TYPE_FORECAST_SERIES:
                await self._forcast_series(series_name, module)
            elif job_type == JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES:
                await self._detect_anomalies(series_name, module)
            else:
                self._analyser_queue.put(
                    {'name': series_name,
                     'error': 'Job type not implemented'})
        else:
            self._analyser_queue.put(
                {'name': series_name,
                 'error': 'Module not implemented'})

    def _handle_response_to_queue(self, response_data):
        if not EnodoJobDataModel.validate_by_job_type(
                response_data, response_data.get('job_type')):
            self._analyser_queue.put(
                {'name': response_data.get('name'),
                 'error': 'Job response not valid'})
            return
        self._analyser_queue.put(response_data)

    async def _analyse_series(self, series_name, analysis_module):
        response = await analysis_module.do_base_analysis()
        self._handle_response_to_queue(
            {'name': series_name,
             'job_type': JOB_TYPE_BASE_SERIES_ANALYSIS, **response})

    async def _check_static_rules(self, series_name, analysis_module):
        response = await analysis_module.do_static_rules_check()
        self._handle_response_to_queue({
            'name': series_name,
            'job_type': JOB_TYPE_STATIC_RULES,
            **response})

    async def _forcast_series(self, series_name, analysis_module):
        """
        Collects data for starting an analysis of a specific time serie
        :param series_name:
        :return:
        """
        error = None
        response = []
        try:
            response = await analysis_module.do_forecast()
        except Exception as e:
            tb = traceback.format_exc()
            error = f"{str(e)}, tb: {tb}"
            logging.error(
                'Error while making and executing forcast module')
            logging.debug(f'Corresponding error: {error}, '
                          f'exception class: {e.__class__.__name__}')
        finally:
            if error is not None:
                self._analyser_queue.put(
                    {'name': series_name,
                     'job_type': JOB_TYPE_FORECAST_SERIES,
                     'error': error})
            else:
                self._handle_response_to_queue(
                    {'name': series_name,
                     'job_type': JOB_TYPE_FORECAST_SERIES,
                     **response})

    async def _detect_anomalies(self, series_name, analysis_module):
        error = None
        response = {}
        try:
            response = await analysis_module.do_anomaly_detect()
        except Exception as e:
            tb = traceback.format_exc()
            error = f"{str(e)}, tb: {tb}"
            logging.error(
                'Error while making and executing anomaly detection module')
            logging.debug(f'Corresponding error: {error}, '
                          f'exception class: {e.__class__.__name__}')
        finally:
            if error is not None:
                self._analyser_queue.put(
                    {'name': series_name,
                     'job_type': JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES,
                     'error': error})
            else:
                self._handle_response_to_queue(
                    {'name': series_name,
                     'job_type': JOB_TYPE_DETECT_ANOMALIES_FOR_SERIES,
                     **response})


async def _save_start_with_timeout(queue, job_data,
                                   siridb_data, siridb_output, modules):
    try:
        analyser = Analyser(queue, siridb_data, siridb_output, modules)
        await analyser.execute_job(job_data)
    except Exception as e:
        tb = traceback.format_exc()
        error = f"{str(e)}, tb: {tb}"
        logging.error('Error while executing Analyzer')
        logging.error(f'Corresponding error: {error}, '
                      f'exception class: {e.__class__.__name__}')


def start_analysing(
        queue, job_data, siridb_data, siridb_output, modules):
    """Switch to new event loop and run forever"""

    try:
        asyncio.run(_save_start_with_timeout(
            queue, job_data, siridb_data, siridb_output, modules))
    except Exception as e:
        logging.error('Error while starting Analyzer')
        logging.debug(f'Corresponding error: {e}, '
                      f'exception class: {e.__class__.__name__}')
        exit()



import socket

HANDSHAKE = 1
HANDSHAKE_OK = 2
HANDSHAKE_FAIL = 3
UNKNOWN_CLIENT = 4
HEARTBEAT = 5
SHUTDOWN = 6
CLIENT_SHUTDOWN = 7

ADD_SERIES = 8
REMOVE_SERIES = 9
LISTENER_ADD_SERIES = 10
LISTENER_REMOVE_SERIES = 11
LISTENER_NEW_SERIES_POINTS = 12
UPDATE_SERIES = 13

RESPONSE_OK = 14

WORKER_JOB = 15
WORKER_JOB_RESULT = 16
WORKER_JOB_CANCEL = 21
WORKER_JOB_CANCELLED = 22
WORKER_UPDATE_BUSY = 23
WORKER_REFUSED = 24


'''
Header:
size,     int,    32bit
type,     int     8bit
packetid, int     8bit

total header length = 48 bits == 6 bytes
'''

PACKET_HEADER_LEN = 6


async def read_packet(sock, header_data=None):
    if isinstance(sock, socket.socket):
        if header_data is None:
            header_data = sock.recv(PACKET_HEADER_LEN)
        body_size, packet_type, packet_id = read_header(header_data)
        return packet_type, packet_id, sock.recv(body_size)
    else:
        if header_data is None:
            header_data = await sock.read(PACKET_HEADER_LEN)
        body_size, packet_type, packet_id = read_header(header_data)
        return packet_type, packet_id, await sock.read(body_size)


def create_header(size, type, id=1):
    return \
        size.to_bytes(4, byteorder='big') + \
        type.to_bytes(1, byteorder='big') + \
        id.to_bytes(1, byteorder='big')


def read_header(binary_data):
    return \
        int.from_bytes(binary_data[:4], 'big'), \
        int.from_bytes(binary_data[4:5], 'big'), \
        int.from_bytes(binary_data[5:6], 'big')




"""
Upload to PyPI
python3 setup.py sdist
twine upload --repository pypitest dist/python-enodo-X.X.X.tar.gz
twine upload --repository pypi dist/python-enodo-X.X.X.tar.gz
"""
from setuptools import setup, find_packages
from enodo import __version__
try:
    with open('README.md', 'r') as f:
        long_description = f.read()
except IOError:
    long_description = '''
The Enodo HUB connector can be used to communicate with the SiriDB Enodo HUB.
    '''.strip()

setup(
    name='python-enodo',
    version=__version__,
    description='SiriDB Enodo Connector',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/SiriDB/siridb-enodo-lib',
    author='Timo Janssen',
    author_email='timo@cesbit.com',
    license='GPLv3',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'func_timeout',
        'numpy',
        'pandas',
        'qpack',
        'siridb-connector',
        'statsmodels',
    ],
    # What does your project relate to?
    keywords='data communication connector enodo siridb library',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['docs', 'tests*']),

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
)



import os
from configparser import RawConfigParser, _UNSET, SectionProxy, ConfigParser

EMPTY_CONFIG_FILE = {
    'worker': {
        'hub_hostname': '',
        'hub_port': '9103',
        'heartbeat_interval': '25',
        'max_job_duration': '120',
        'internal_security_token': ''
    },
    'siridb_data': {
        'host': '',
        'port': '',
        'user': '',
        'password': '',
        'database': '',
    },
    'siridb_output': {
        'host': '',
        'port': '',
        'user': '',
        'password': '',
        'database': '',
    }
}


def create_standard_config_file(path):
    _config = ConfigParser()

    for section in EMPTY_CONFIG_FILE:
        _config.add_section(section)
        for option in EMPTY_CONFIG_FILE[section]:
            _config.set(section, option,
                        EMPTY_CONFIG_FILE[section][option])

    with open(path, "w") as fh:
        _config.write(fh)


class EnodoConfigParser(RawConfigParser):

    def __getitem__(self, key):
        if key != self.default_section and not self.has_section(key):
            return SectionProxy(self, key)
        return self._proxies[key]

    def has_option(self, section, option):
        return True

    def get(
            self, section, option, *, raw=False, vars=None,
            fallback=_UNSET):
        """Edited default get func from RawConfigParser
        """
        env_value = os.getenv(
            f"ENODO_{section.upper()}_{option.upper()}")
        if env_value is not None:
            return env_value

        try:
            return super(EnodoConfigParser, self).get(
                section, option, raw=False, vars=None, fallback=_UNSET)
        except Exception as _:
            raise Exception(
                f'Invalid config, missing option "{option}" in section '
                f'{section}" or environment variable '
                f'"ENODO_{section.upper()}_{option.upper()}"')
